/********************************************************
 * line_number -- demonstrate a line_number class.	*
 *							*
 * Programming Problem: 14-4.				*
 ********************************************************/
#include <iostream.h>
#include "test_txt.h"	/* Get the test data */

/********************************************************
 * line_number -- A line number class			*
 *							*
 * Maintains a character position within a text buffer	*
 * that represents a line number.			*
 *							*
 * Member functions:					*
 *	goto_line(int line) -- Position to a line number*
 *	int get_current_line() -- return current line	*
 *	unsigned long int get_char_pos() 		*
 *		-- get character index			*
 *							*
 * Note: Line numbers start with 1.			*
 *	Character position starts with 0.		*
 ********************************************************/
class line_number {
    private:
	unsigned int cur_line;		// Line we are at now
	unsigned long int cur_pos;	// Character index

    public:
	line_number(void) {
	    cur_line = 1;
	    cur_pos = 0;
	}
	// Destructor defaults
    private:
	// No copy constructor
	line_number(const line_number &);	

	// No assignment operator
	line_number &operator = (const line_number &);
    
    public:
	// Goto a given line
	void goto_line(const unsigned int line);	

	// Return the current line number
	unsigned int get_current_line(void) {
	    return (cur_line);
	}

	unsigned long int get_char_pos(void) {
	    return (cur_pos);
	}

    private:
	friend void find_nearest_pos(
	    const unsigned int seek_line,
	    unsigned int got_line,
	    unsigned int got_pos);
};


// A set of line numbers
const int LIST_SIZE = 20;	// Number of line number classes
class line_number line_number_list[LIST_SIZE];

/********************************************************
 * line_number::goto_line -- position a line number	*
 *		class at a given line number		*
 ********************************************************/
void line_number::goto_line(
   const unsigned int line	// Line number we want to go to
)
{
    find_nearest_pos(line, cur_line, cur_pos);

    while ((line > cur_line) && (text[cur_pos] != '\0')) {
	if (text[cur_pos] == '\n') {
	    ++cur_line;
	}
	++cur_pos;
    }
    while ((line < cur_line) && (cur_pos > 0)) {
	if (text[cur_pos] == '\n') {
	    --cur_line;
	}
	--cur_pos;
    }
    // Move to beginning of the line
    while ((cur_pos > 0) && (text[cur_pos-1] != '\n'))
	--cur_pos;
}


/********************************************************
 * abs -- return the absolute value of an integer	*
 ********************************************************/
inline int abs(int value) {
    if (value >= 0)
	return (value);
    return (-value);
}
/********************************************************
 * find_nearest_pos -- find a line number class		*
 *	that is nearest the line we want and return	*
 *	the position for that class.			*
 ********************************************************/
void find_nearest_pos(
    const unsigned int seek_line,	// Line we want
    unsigned int got_line,		// Line we got
    unsigned int got_pos		// Position we got
)
{
    int i;	// Index into line number list
    int close_index; // Closest line number
    int close_delta;  // Difference of closest line

    close_index = 0;
    close_delta = abs(
	line_number_list[close_index].get_current_line() - seek_line);

    for (i = 0; i < LIST_SIZE; ++i) {
	// Difference between what we want and what we have
	int check_delta = 
	    abs(line_number_list[i].get_current_line() - seek_line);

	if (check_delta < close_delta) {
	    close_delta = check_delta;
	    close_index = i;
	}
	if (close_delta == 0)
	    break;
    }
    got_line = line_number_list[i].get_current_line();
    got_pos = line_number_list[i].get_char_pos();
}


/********************************************************
 * dump -- write out the first few characters of a line	*
 ********************************************************/
void dump(
    unsigned int index	// Index of the line to dump
)
{
    int i;		// Character count
    unsigned long int cur_pos;	// Position we are dumping

    cout << "Line #" << index << ": ";

    cur_pos = line_number_list[index].get_char_pos();

    for (i = 0; i < 30; ++i) {
        if (text[cur_pos] == '\n')
	    break;
	cout << text[cur_pos];
	++cur_pos;
    }
    cout << '\n';
}
main()
{
    line_number_list[1].goto_line(1);
    line_number_list[3].goto_line(3);
    line_number_list[5].goto_line(5);
    line_number_list[2].goto_line(2);
    line_number_list[4].goto_line(4);
    line_number_list[10].goto_line(10);
    line_number_list[9].goto_line(9);
    line_number_list[11].goto_line(11);

    dump(1);
    dump(2);
    dump(3);
    dump(4);
    dump(5);
    dump(9);
    dump(10);
    dump(11);
    return (0);
}
