/********************************************************
 * sparse -- a class to implement sparse array.		*
 *	A sparse array has only a few non-zero elements.*
 *							*
 * Programming Problem: 18-3				*
 ********************************************************/
#include <iostream.h>
/********************************************************
 * class sparse -- sparse array.			*
 *							*
 * Member functions:					*
 *	const []	-- Return an element of the 	*
 *				array			*
 *	a = a + a	-- Add two arrays together	*
 *	a = a - a	-- Subtract two arrays.		*
 *	a = a		-- Assignment operator.		*
 *	a == a		-- Two arrays are equal		*
 *	a != a		-- Two arrays are not equal	*
 *	set(index, value) -- Set a value in the array	*
 *							*
 * Note: We do not implement multiplication or 		*
 * 	division because these can be confusing.	*
 *	Is multiplication dot product, cross product	*
 *	or element by element multiplication?		*
 *	Better to defined named functions to handle	*
 *	these operations.				*
 ********************************************************/
const int SPARSE_MAX = 100;	// Max number of non-zero elements in an array
const int FREE_INDEX = -1;	// Index of an element that's free

class sparse {
    private:
	/* 
	 * Structure to hold elements of the array
	 */
	struct array_element {
	    int index;		// Index of the element 
	    int value;		// Value of the element
	};

	// Place we store the data
	struct array_element data[SPARSE_MAX];

    public:
        sparse(void);	// Default constructor
	sparse(const sparse &old_sparse);	// Copy constructor
	// Destructor defaults
	sparse operator = (const sparse &old_sparse);

	const int &operator[](const int index) const;

	sparse operator +(const sparse &a2);
	sparse operator -(const sparse &a2);
	int operator ==(const sparse &a2);
	int operator !=(const sparse &a2);

	void set(const int index, const int value);
    private:
	int find_element(
	    const int index	// Index of the element we want to find
	) const;
};
/********************************************************
 * sparse::sparse -- create a spares array		*
 ********************************************************/
sparse::sparse(void)
{
    int i;	// Index into the data array

    for (i = 0; i < SPARSE_MAX; ++i)
        data[i].index = FREE_INDEX;
}
/********************************************************
 * sparse::sparse(const sparse &) -- copy constructor	*
 ********************************************************/
sparse::sparse(
    const sparse &old_sparse	// Sparse array we will copy
)
{
    int i;	// Index into the data array

    // Note: memcpy would be more efficient, but less instructive
    for (i = 0; i < SPARSE_MAX; ++i) {
        if (old_sparse.data[i].index == FREE_INDEX)
	    data[i].index = FREE_INDEX;
	else
	    data[i] = old_sparse.data[i];
    }
}
/********************************************************
 * sparse::operator =  -- copy a sparse array		*
 ********************************************************/
sparse sparse::operator = (const sparse &old_sparse)
{
    int i;	// Index into the data array

    // Note: memcpy would be more efficient, but less instructive
    for (i = 0; i < SPARSE_MAX; ++i) {
        if (old_sparse.data[i].index == FREE_INDEX)
	    data[i].index = FREE_INDEX;
	else
	    data[i] = old_sparse.data[i];
    }
    return (*this);
}
/********************************************************
 * sparse::find_element -- find an element in the 	*
 *		sparse array.				*
 *							*
 * Returns index into the data or FREE_INDEX if the 	*
 * data is not in the array.				*
 ********************************************************/
int sparse::find_element(
    const int index	// Index of the element we want to find
) const 
{
    int i;	// Index of the item we want to find

    for (i = 0; i < SPARSE_MAX; ++i) {
        if (data[i].index == index)
	    return (i);
    }
    return (FREE_INDEX);
}
/********************************************************
 * sparse[] -- return an element from the array		*
 ********************************************************/
const int & sparse::operator[](const int index) const {
    int i;	// Index into the data

    i = find_element(index);
    if (i != FREE_INDEX)
	return (data[i].value);

    static int zero = 0;
    return (zero);
}
/********************************************************
 * sparse::set -- set an element of the array		*
 ********************************************************/
void sparse::set(
    const int index,	// Index into the array
    const int value	// Data for that index
) {
    int i;	// Index into the array

    for (i = 0; i < SPARSE_MAX; ++i) {
        if (data[i].index == index) {
	    data[i].value = value;
	    return;
	}
    }

    // Didn't find an existing cell, try to find a free one
    for (i = 0; i < SPARSE_MAX; ++i) {
        if (data[i].index == FREE_INDEX) {
	    data[i].index = index;
	    data[i].value = value;
	    return;
	}
    }
    cout << "Error: Ran out of room in sparse matrix\n";
}
/********************************************************
 * sparse + sparse -- add two sparse arrays		*
 ********************************************************/
sparse sparse::operator +(const sparse &a2) {
    int i;	// Index into the data (array 1)
    sparse result(*this);	// Result

    for (i = 0; i < SPARSE_MAX; ++i) {
	if (a2.data[i].index != FREE_INDEX) {
	    int index;	// Element index

	    // Look for index in current array
	    index = result.find_element(a2.data[i].index);
	    if (index == FREE_INDEX)
	        result.set(a2.data[i].index, a2.data[i].value);
	    else
	        result.data[index].value += a2.data[i].value;
	}
    }
    return (result);
}
/********************************************************
 * sparse - sparse -- subtract two sparse arrays	*
 ********************************************************/
sparse sparse::operator -(const sparse &a2) {
    int i;	// Index into the data (array 1)
    sparse result(*this);	// Result

    for (i = 0; i < SPARSE_MAX; ++i) {
	if (a2.data[i].index != FREE_INDEX) {
	    int index;	// Element index

	    // Look for index in current array
	    index = result.find_element(a2.data[i].index);
	    if (index == FREE_INDEX)
	        result.set(a2.data[i].index, -a2.data[i].value);
	    else
	        result.data[index].value -= a2.data[i].value;
	}
    }
    return (result);
}
/********************************************************
 * sparse == sparse -- see if two sparse arrays are =	*
 ********************************************************/
int sparse::operator ==(const sparse &a2) {
    int i;	// Index into the data (array 1)
    int count_1 = 0;// Number of non-zero elements in array 1
    int count_2 = 0;// Number of non-zero elements in array 2

    for (i = 0; i < SPARSE_MAX; ++i) {
        if (data[i].index != FREE_INDEX)
	    ++count_1;
    }
    for (i = 0; i < SPARSE_MAX; ++i) {
        if (a2.data[i].index != FREE_INDEX)
	    ++count_2;
    }

    if (count_1 != count_2)
        return (0);	// Counts not the same, not equal

    for (i = 0; i < SPARSE_MAX; ++i) {
	if (a2.data[i].index != FREE_INDEX) {
	    int index;	// Element index

	    // Look for index in current array
	    index = find_element(a2.data[i].index);
	    if (index == FREE_INDEX)
	        return (0);		// Different elements
	    else {
	        if (data[index].value != a2.data[i].value)
		    return (0);		// Different data
	    }
	}
    }
    return (1);		// Everything matches
}
/********************************************************
 * sparse != sparse -- see if two sparse arrays are !=	*
 ********************************************************/
int sparse::operator !=(const sparse &a2) {
    return (!(*this == a2));
}


const unsigned int TEST_MAX = 1000;	// Biggest index in a test array
/********************************************************
 * print_array -- print an array to cout.		*
 ********************************************************/
void print_array(
    const sparse &array	// array to print
) {
    unsigned int i;	// index for printing

    for (i = 0; i < TEST_MAX; ++i) {
        int value = array[i];
	if (value != 0)
	    cout << '[' << i << "]=" << value << ' ';
    }
}

main()
{ 
    sparse s1, s2;	// Two sparse array
    sparse result;	// Result of the last operation

    s1.set(1,1);
    s1.set(99, 99);
    s1.set(400, 400);

    cout << "s1 = ";
    print_array(s1);
    cout << '\n';

    s2.set(2,2);
    s2.set(99, 99);
    s2.set(405, 405);

    cout << "s2 = ";
    print_array(s2);
    cout << '\n';

    result = s1 + s2;
    cout << "s1 + s2 = ";
    print_array(result);
    cout << '\n';

    result = s1 - s2;
    cout << "s1 - s2 = ";
    print_array(result);
    cout << '\n';

    cout << "s1 == s1 is " << (s1 == s1) << '\n';
    cout << "s1 == s2 is " << (s1 == s2) << '\n';
    cout << "s1 != s1 is " << (s1 != s1) << '\n';
    cout << "s1 != s2 is " << (s1 != s2) << '\n';

    return (0);
}
