/********************************************************
 * Float -- Demonstrate how floating point is done.	*
 *							*
 * Programming Problem: 19-1				*
 *							*
 * This program implements and uses a class that 	*
 * defines a primitive floating point number type.	*
 * Data is stored as ASCII and manipulated in ASCII.	*
 * The idea here is not speed, but instead we want	*
 * to demonstrate the steps that go into floating	*
 * point calculations.					*
 *							*
 * Warning: The programmer who wrote this is not an	*
 * expert in computer architecture.  The algorithms 	*
 * presented here may not be the best.			*
 *							*
 * I'm sure someone out there who knows what's he's	*
 * doing can tell me how to do this better.  Send me	*
 * E-mail at sdo@crash.cts.com				*
 ********************************************************/
#include <iostream.h>
#include <stdlib.h>

const int FRACT_SIZE = 4;	// Size of a fraction
class sfloat {
    public:
	/*
	 * Numbers have the form
	 *   +X.XXXe+Y
	 *   ^^^^^^ ^+---- exponent
	 *   ^^^^^^ +----- exponent_sign
	 *   ^+++++------- fraction
	 *   +------------ sign
	 */
	char sign;		
	char fraction[FRACT_SIZE];
	char exponent_sign;
	char exponent;

    public:
	sfloat(void);
	// Destructor defaults
	// Copy constructor defaults
	// Assignment operator defaults
	sfloat(const char string[]);
    public:
	// Add one to the exponent
        void inc_exponent(void) {
	    if (exponent_sign == '-') {
		 --exponent;
	         if (exponent <= '0') {
	            exponent_sign = '+';
		    exponent = '0';
		}
	    } else
	       ++exponent;
	}
	// Subtract one from the exponent
	void dec_exponent(void) {
	    if (exponent_sign == '+') {
		 --exponent;
	         if (exponent < '0') {
	            exponent_sign = '-';
		    exponent = '1';
		}
	    } else
	       --exponent;
	}
	// Return the exponent as an integer
	int exponent_value(void) {
	    if (exponent_sign == '-')
	        return (- (exponent - '0'));
	    return (+ (exponent - '0'));
	}

};
/********************************************************
 * sfloat::sfloat -- initialize a simple floating point	*
 *	number.						*
 ********************************************************/
sfloat::sfloat(void)
{
    int i;	// General purpose index
    sign = '+';
    for (i = 0; i < FRACT_SIZE; ++i) {
	fraction[i] = '0';
    }
    exponent_sign = '+';
    exponent = '0';
}
/*
 *	012345678
 * 	+X.XXXe+Y
 */
/********************************************************
 * sfloat::sfloat -- initialize a simple floating point	*
 *	number.						*
 ********************************************************/
sfloat::sfloat(
    const char string[]	// String version of the float #
) {
    int i;	// General purpose index
    sign = string[0];

    fraction[0] = string[1];
    for (i = 1; i < FRACT_SIZE; ++i) {
	fraction[i] = string[2+i];
    }
    exponent_sign = string[FRACT_SIZE+3];
    exponent = string[FRACT_SIZE+4];
}
/********************************************************
 * shift the fraction to the right one.			*
 ********************************************************/
void shift_right(
    char data[],		// Data to shift
    int size			// Number of bytes in data
) {
    int i;			// Character index

    for (i = size-1; i > 0; --i)
       data[i] = data[i-1];
    data[0] = '0';
}
/********************************************************
 * shift the fraction to the left one.			*
 ********************************************************/
void shift_left(
    char data[],		// Data to shift
    int size			// Number of bytes in data
) {
    int i;			// Character index

    for (i = 0; i < size-1; ++i)
       data[i] = data[i+1];
    data[size-1] = '0';
}
/********************************************************
 * normalize_and_copy -- Copy a fraction from		*
 *	a working copy into the result, normalizing	*
 *	the number in the process			*
 *							*
 * Note: This algorithm takes into account that the	*
 *	implied decimal point is digit 2 of the fraction*
 *	(fraction=XX.XXX) and adjusts things accordingly*
 ********************************************************/
void normalize_and_copy(
    sfloat &result,		// Place to put the result
    char result_fraction[FRACT_SIZE+2]	// Fraction to copy
)
{
    int i;	// Random index
    if (result_fraction[0] != '0') {
	shift_right(result_fraction, sizeof(result_fraction));
	result.inc_exponent();
    }
    // Check for 0.0
    for (i = 0; i < FRACT_SIZE+1; ++i) {
        if (result_fraction[i] != '0') 
	    goto not_zero;
    }
    // We have a zero
    result.sign = '+';
    for (i = 0; i < FRACT_SIZE; ++i)
       result.fraction[i] = '0';
    result.exponent_sign = '+';
    result.exponent = '0';
    return;

not_zero:
    while (result_fraction[1] == '0') {
	shift_left(&result_fraction[1], sizeof(result_fraction)-1);
	result.dec_exponent();
    }
    for (i = 0; i < FRACT_SIZE; ++i)
       result.fraction[i] = result_fraction[i+1];
}

/********************************************************
 * copy_and_guard -- copy the fraction to a working	*
 *	fraction and add a guard digit.			*
 ********************************************************/
static void copy_and_guard(
    char fract[FRACT_SIZE+1],	// Fraction we copy into
    const sfloat &number	// Number we copy out of
)
{
    int i;		// Fraction digit index

    for (i = 0; i < FRACT_SIZE; ++i)
        fract[i] = number.fraction[i];
    fract[FRACT_SIZE] = '0';		// Add guard digit
}

/********************************************************
 * compare_fraction -- Compare two fractions		*
 *							*
 * Returns						*
 *	0 -- fractions are equal			*
 *	<0  -- f1 < f2					*
 *	>0  -- f1 > f2					*
 ********************************************************/
int compare_fraction(
    const char f1[],	// Fraction #1
    const char f2[]	// Fraction #2
) {
    int i;	// Fraction digit indexer

    for (i = 0; i < FRACT_SIZE; ++i) {
        // Result of the difference in fraction digits
        int result = f1[i] - f2[i];

        if (result != 0)
            return (result);
    }
    return (0);
}
/********************************************************
 * subtract_fraction -- subtract two fractions.		*
 *	Subtracts the larger fraction from the smaller	*
 *							*
 * Note: Caller is expected to fixup the sign		*
 ********************************************************/
void subtract_fraction(
    char result_fract[],	// Place to put result
    const char s1_fract[],	// Fraction for part 1
    const char s2_fract[]	// Fraction for part 2
)
{
    const char *s1_ptr = s1_fract;	// Pointer fraction 1
    const char *s2_ptr = s2_fract;	// Pointer fraction 2
    int borrow = 0;		// Value of the borrow between digits
    int i;			// General purpose index

    if (compare_fraction(s1_ptr, s2_ptr) < 0) {
        s1_ptr = s2_fract;
	s2_ptr = s1_fract;
    }
    for (i = FRACT_SIZE; i >= 0; --i) {
	// Digit resulting from this add
	int digit = (s1_ptr[i] -'0') - (s2_ptr[i]  -'0') - borrow;
	if (digit < 0) {
	    borrow = 1;
	    digit += 10;
	} else
	    borrow = 0;

	result_fract[i+1] = digit + '0';	// Turn back into ascii
    }

    if (borrow != 0) {
        cerr << "Internal error: Unexpected borrow\n";
	exit (8);
    }
    result_fract[0] = '0';
}
/********************************************************
 * Add two fractions.  The signs of the fractions	*
 *	should be the same.				*
 ********************************************************/
void add_fraction(
    char result_fract[],	// Place to put result
    const char s1_fract[],	// Fraction for part 1
    const char s2_fract[]	// Fraction for part 2
)
{
    int carry = 0;		// Value of the carry between digits
    int i;			// General purpose index
    for (i = FRACT_SIZE; i >= 0; --i) {
	// Digit resulting from this add
	int digit = (s1_fract[i] -'0') + (s2_fract[i]  -'0')+ carry;
	if (digit >= 10) {
	    carry = 1;
	    digit %= 10;
	} else
	    carry = 0;

	result_fract[i+1] = digit + '0';	// Turn back into ascii
    }

    result_fract[0] = carry + '0';
}
/********************************************************
 * sfloat + sfloat -- add two simple floating numbers	*
 *	together.					*
 *							*
 * Warning: This is not as simple as it sounds.		*
 ********************************************************/
sfloat operator +(
    const sfloat &s1, 	// Take this number and
    const sfloat &s2	// add it to this number
) {
    sfloat s1_work(s1);			// Working version of s1
    sfloat s2_work(s2);			// Working version of s2
    char s1_fract[FRACT_SIZE+1];	// Working fraction for s1
    char s2_fract[FRACT_SIZE+1];	// Working fraction for s2

    /*
     * Copy the fraction to the working area and
     * put on the guard digit.
     * STEP 2.  Page 344
     */
    copy_and_guard(s1_fract, s1_work);
    copy_and_guard(s2_fract, s2_work);
    /* 
     * Shift the smallest exponent to the right and
     * increment the exponent.  Continue until exponents
     * match.
     * STEP 3. Page 344
     */
    // Value of the exponents
    int s1_exponent = s1_work.exponent_value();
    int s2_exponent = s2_work.exponent_value();
    while (s1_exponent != s2_exponent) {
        if (s1_exponent > s2_exponent) {
	    shift_right(s2_fract, sizeof(s2_fract));
	    s2_work.inc_exponent();
	    ++s2_exponent;
	} else {
	    shift_right(s1_fract, sizeof(s1_fract));
	    s1_work.inc_exponent();
	    ++s1_exponent;
	}
    }
    /*
     * Add the two fractions.
     * Step 4, Page 345
     */
    // Fraction for result of the add
    // (Two digits are before the implied decimal point)
    // We use FRACT_SIZE+2 (1 on the left for overflow)
    // (one on the right for guard)
    char result_fract[FRACT_SIZE+2];

    if (s1_work.sign == s2_work.sign)
        add_fraction(result_fract, s1_fract, s2_fract);
    else 
        subtract_fraction(result_fract, s1_fract, s2_fract);
        
    int i;		// General purpose integer
    /*
     * If guard digit > 5 round result
     */
    if (result_fract[FRACT_SIZE+1] >= '5') {
	// The carry is one for the last digit
	int carry = 1;

	for (i = FRACT_SIZE+1; i >= 0; --i) {
	    // Add in the carry
	    result_fract[i] += carry;

	    // Did this digit overflow?
	    if (result_fract[i] > '9') {
		// It did, we carry into the next digit
		carry = 1;
		result_fract[i] -= 10;
	    } else
		carry = 0;
	}
    }

    sfloat result;	// Result of all this work

    result.sign = s1_work.sign;
    if (s1_work.sign != s2_work.sign) {
        if (compare_fraction(s1_fract, s2_fract) < 0)
	    result.sign = s2_work.sign;
    }
    result.exponent = s1_work.exponent;
    result.exponent_sign = s1_work.exponent_sign;

    /*
     * Normalize the result
     */
    normalize_and_copy(result, result_fract);
    return (result);
}
/********************************************************
 * subtract two sfloats -- just a sign change and an add*
 ********************************************************/
sfloat operator -(
    const sfloat &s1, 	// Take this number and
    const sfloat &s2	// add it to this number
) {
    sfloat s1_work(s1);			// Working version of s1
    switch (s1_work.sign) {
        case '+':
	    s1_work.sign = '-';
	    break;
	case '-':
	    s1_work.sign = '+';
	    break;
	default:
	    cerr << "Internal Error: Illegal sign in subtract\n";
	    exit (8);
    }
    return (s1_work + s2);
}
/********************************************************
 * << sfloat -- output a simple floating point number	*
 ********************************************************/
ostream & operator << (
    ostream &out_file,	// File to write to
    const sfloat number	// Number to write
) {
    int i;	// General purpose index
    out_file << number.sign;
    out_file << number.fraction[0];
    out_file << '.';
    for (i = 1; i < FRACT_SIZE; ++i)
	out_file << number.fraction[i];

    out_file << 'E';
    out_file << number.exponent_sign;
    out_file << number.exponent;
    return (out_file);
};

/********************************************************
 * add_test -- add two sfloat numbers and display the	*
 *		results.				*
 ********************************************************/
void add_test(
    const char str1[],	// String versions of our numbers
    const char str2[]	// String version of number 2
)
{   
    // Two numbers to add
    sfloat n1(str1), n2(str2);

    cout << n1 << " + " << n2 << " = " << (n1+n2) << '\n';
}
/********************************************************
 * sub_test -- subtract two sfloat numbers and display 	*
 *	the results.					*
 ********************************************************/
void sub_test(
    const char str1[],	// String versions of our numbers
    const char str2[]	// String version of number 2
)
{   
    // Two numbers to add
    sfloat n1(str1), n2(str2);

    cout << n1 << " - " << n2 << " = " << (n1-n2) << '\n';
}
main()
{
    // Test basic addition
    add_test("+1.000E+0", "+1.000E+0");

    // Test when we have a carry
    add_test("+9.000E+0", "+9.000E+0");

    // Test shifting
    add_test("+1.000E+0", "+1.000E-1");

    // Rounding test
    cout << '\n';
    add_test("+1.000E+0", "+9.000E-1");
    add_test("+1.000E+0", "+9.000E-2");
    add_test("+1.000E+0", "+9.000E-3");
    add_test("+1.000E+0", "+9.000E-4");
    add_test("+1.000E+0", "+9.000E-5");

    // Rounding test
    cout << '\n';
    add_test("+1.000E+0", "+9.999E-1");
    add_test("+1.000E+0", "+9.999E-2");
    add_test("+1.000E+0", "+9.999E-3");
    add_test("+1.000E+0", "+9.999E-4");
    add_test("+1.000E+0", "+9.999E-5");

    // Check things out when we have sign trouble
    cout << '\n';
    add_test("+1.000E+0", "-1.000E+0");
    add_test("-1.000E+0", "+1.000E+0");
    add_test("+2.000E+0", "-1.000E+0");
    add_test("-2.000E+0", "+1.000E+0");
    add_test("+1.000E+0", "-2.000E+0");
    add_test("-1.000E+0", "+2.000E+0");

    /*
     * Subtract tests
     */
    cout << '\n';
    sub_test("+1.000E+0", "+1.000E+0");
    sub_test("+9.000E+0", "+9.000E+0");
    sub_test("+1.000E+0", "+1.000E-1");
    return (0);
}


 

