/********************************************************
 * print -- a page formatting class			*
 *							*
 * Prints a page of the form:				*
 *	+-------------------------------+		*
 *	| Heading 			|		*
 *	| -- blank line--		|		*
 *	| Line 1			|		*
 *	| Line 2			|		*
 *	| Line 3			|		*
 *	| Line 4			|		*
 *	| ......			|		*
 *	| Line 50			|		*
 *	| -- blank line--		|		*
 *	| Page: nn			|		*
 *	+-------------------------------+		*
 *							*
 * Member functions:					*
 *	open_file(name)	-- Opens the file		*
 *	define_header(header) -- Sets the header	*
 *			(optional)			*
 *	print_line(line) -- Output a line		*
 *	page(void) -- Finish a page			*
 *	close_file(void) -- close the file.		*
 ********************************************************/
#ifndef _PRINT_H_	// Prevent double includes
#define _PRINT_H_

const int PAGE_SIZE = 50;	// Number of text lines per page
				// (Excluding header and footer)
class print {
    protected:
	ofstream out_file;	// Output file
	const char *header_ptr;	// Pointer to the header
	int page_number;	// Current page number

    private:
	int cur_line;		// Current line number
    
    public:
	print(void) { 		// Default constructor
	   cur_line = 0;
	   page_number = 1;
	   header_ptr = "";
	}		
	virtual ~print(){}	// Destructor
    private:
	print(const print &);	// No copy constructor
	print operator =(const print &); // No assignment operator
    
    public:
	// Open the output file
	// Returns bad flag
	int open_file(
	    const char name[]	// Name of the file to open
	) {
	    out_file.open(name);
	    return (out_file.bad());
	}

	// Close the output file
	void close_file(void) {
	    out_file.close();
	}
	    
	// Print the header
	virtual void print_header(void) {
	    out_file << header_ptr << '\n';
	    out_file << '\n';
	}

	// Print the footer
	virtual void print_footer(void) {
	    out_file << '\n';
	    out_file << "Page: " << page_number << '\n';
	    ++page_number;
	    out_file << '\f';
	}

	// Initialize the header 
	void define_header(
	    const char *const new_head	// New version of the header
	) {
	    header_ptr = new_head;
	}

	// Print a single line
	void print_line(
	    const char line[]	// Line to print
	) {
	    if (cur_line == 0) {
		print_header();
	    } 

	    out_file << line << '\n';
	    ++cur_line;

	    if (cur_line == PAGE_SIZE) {
		page();
	    }
	}

        // Output a page
	void page(void) {
	    print_footer();
	    cur_line = 0;
	}
};

#endif /* _PRINT_H_ */	// Prevent double includes
