/********************************************************
 * split -- function to find good split points in a line*
 ********************************************************/
#include <ctype.h>
#include <string.h>
#include "split.h"

/********************************************************
 * split -- returns a good split point in a line	*
 *	or -1 if the line is too short to split		*
 ********************************************************/
int split(
    const char line[],	// Line to split
    const int point	// Point at or before the break
)
{
    // Too short to split
    if (strlen(line) < point)
	return (-1);

    // Possible split point
    register const char *split_point;

    for (split_point = &line[point]; split_point > line; --split_point) {
	if (isspace(*split_point)) 
	    return (split_point - line);
    }

    // Couldn't find a split point with spaces.  Try to find one that's 
    // a symbol
    for (split_point = &line[point]; split_point > line; --split_point) {
	if (!isalnum(*split_point)) 
	    return (split_point - line);
    }

    // Couldn't find anything good.  Try non-letter
    for (split_point = &line[point]; split_point > line; --split_point) {
	if (!isalpha(*split_point)) 
	    return (split_point - line);
    }

    // Very long line throw in the towel.
    return (point);
}
